/*
 * add gnu
 * gcc -O -o latency latency.c
 *
 * latency low-bytes high-bytes [stride-bytes]
 *
 * If stride-bytes is supplied, the program prints out the
 * number of nanoseconds per iteration of this loop for sizes
 * between low-bytes and high-bytes:
 *
 * char a[size];
 * for(i = 0; i < size; i += stride)
 *   junk = a[i];
 *
 * That is, the program prints out the cost of a load instruction.
 * The sequential access avoids persistent misses in small TLBs.
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <setjmp.h>
#ifdef __svr4__
#include <sys/systeminfo.h>
#endif

jmp_buf jb;
volatile int done;
void pagecatchalrm(int);
void cachecatchalrm(int);
int pageloop(int size);
int readnum(char *s);

int
main(int argc, char *argv[])
{
  int stride; /* bytes between successive loads */
  int low;    /* start by using this many bytes of memory */
  int high;   /* end before using this much */
  int paces;  /* how many strides to load before starting at the beginning */
  int sz;
  struct rlimit rl;
  char sibuf[512];

  if(argc != 4){
    fprintf(stderr, "Usage: latency low-bytes high-bytes [stride-bytes]\n");
    exit(1);
  }

  low = readnum(argv[1]);
  high = readnum(argv[2]);
  if(argc > 3){
    stride = readnum(argv[3]);
    if(stride < sizeof(int*)){
      fprintf(stderr, "latency: the stride must be >= %d\n", sizeof(int*));
      exit(1);
    }
    if((stride % sizeof(int*)) != 0){
      fprintf(stderr, "latency: the stride must be a multiple of %d\n",
              sizeof(int*));
      exit(1);
    }
  } else {
    stride = 0;
  }
  if(low < stride){
    fprintf(stderr, "latency: low-bytes must be >= stride\n");
    exit(1);
  }
  if(high < low + stride){
    fprintf(stderr, "latency: high-bytes must be >= low-bytes + stride\n");
    exit(1);
  }
  if(low < 1024){
    fprintf(stderr, "latency: low must be >= 1024\n");
    exit(1);
  }

  rl.rlim_cur = rl.rlim_max = 0;
  getrlimit(RLIMIT_DATA, &rl);
  rl.rlim_cur = rl.rlim_max;
  setrlimit(RLIMIT_DATA, &rl);
  rl.rlim_cur = rl.rlim_max = 0;
  getrlimit(RLIMIT_DATA, &rl);

  sibuf[0] = '\0';
#ifdef __svr4__
  sysinfo(SI_PLATFORM, sibuf, sizeof(sibuf)-1);
#endif
  printf("# low=%d high=%d stride=%d ", low, high, stride);
  printf("rlim_cur=%d ", rl.rlim_cur);
  printf("%s ", sibuf);
  printf("\n");
  fflush(stdout);

  for(sz = low; sz <= high; sz *= 1.1){
    if(stride){
      paces = sz / stride;
      cacheloop(stride, paces);
    } else {
      pageloop(sz);
    }
  }

  exit(0);
}

int
cacheloop(int stride, int paces)
{
  long long stmp;
  int size;
  int istride = stride / sizeof(int*);
  char *mem;
  int **ip;
  int iters;
  int loads;
  double t;
  struct timeval tv0, tv1;

  stmp = stride * (long long) paces;
  if(stmp > 2000000000 || stmp <= 0){
    fprintf(stderr, "latency: total size too big\n");
    exit(1);
  }
  size = stmp;
  mem = (char *) malloc(size);
  if(mem == 0){
    fprintf(stderr, "latency: malloc(%d) failed\n", size);
    exit(1);
  }

  /*
   * Create a linked list of cache blocks. Following this list
   * forces the CPU to wait for each load before it starts
   * the next. Idea from Lance Berc.
   */
  memset(mem, '\0', size);
  for(ip = (int **) mem; ip < (int **) (mem + size - stride); ip += istride){
    *ip = (int*) (ip + istride);
  }
  
  done = 0;
  signal(SIGALRM, cachecatchalrm);
  alarm(3);
  gettimeofday(&tv0, (struct timezone *) 0);


  for(iters = 0; done == 0; iters++){
    ip = (int **) mem;

    /*
     * This is the inner loop that we're timing.
     *
     * gcc -O on an x86 generates this code:
     * L44:
     *  movl (%edx),%edx
     *  testl %edx,%edx
     *  jne L44
     *
     * gcc -O on a SPARC:
     * L42:
     *  cmp %o1,0
     *  bne,a L42
     *  ld [%o1],%o1
     */
    while(ip)
      ip = (int **) *ip;
  }

  gettimeofday(&tv1, (struct timezone *) 0);
  alarm(0);

  t = tv1.tv_sec - tv0.tv_sec;
  t += (tv1.tv_usec - tv0.tv_usec) / 1000000.0;

  loads = iters * paces;
  printf("%d %.1f\n", size, (t / loads) * 1000000000.0);
  fflush(stdout);
  fsync(1);

  free(mem);
}

/*
 * This version is intended to examine paging behavior.
 * It touches randomly selected pages.
 * It essentially ignores the stride.
 * 
 * char a[size];
 * while(1)
 *   junk = a[random() % size];
 *
 * The randomization avoids artifacts caused by the O/S pre-fetching
 * multi-page clusters. Note that the stride is basically ignored.
 */
int
pageloop(int size)
{
  volatile int sum;
  char *mem;
  volatile int loads;
  double t;
  struct timeval tv0, tv1;

  mem = (char *) malloc(size);
  if(mem == 0){
    fprintf(stderr, "latency: malloc(%d) failed\n", size);
    exit(1);
  }
  memset(mem, '\0', size);

  loads = 0;
  sum = 0;
  if(setjmp(jb) == 0){
    signal(SIGALRM, pagecatchalrm);
    alarm(3);
    gettimeofday(&tv0, (struct timezone *) 0);
    while(1){
      char *p = mem + (random() % size);
      sum += *p;
      loads++;
    }
  }

  gettimeofday(&tv1, (struct timezone *) 0);
  alarm(0);

  t = tv1.tv_sec - tv0.tv_sec;
  t += (tv1.tv_usec - tv0.tv_usec) / 1000000.0;

  printf("%d %.1f\n", size, (t / loads) * 1000000000.0);
  fflush(stdout);
  fsync(1);

  free(mem);

  /*
   * If sum is never used, gcc optimizes away the load in the
   * inner loop above.
   */
  return(sum);
}

void
pagecatchalrm(int junk)
{
  longjmp(jb, 1);
}

void
cachecatchalrm(int junk)
{
  done = 1;
}

int
readnum(char *s)
{
  int x, i;
  long long ret;

  if(!isdigit(s[0])){
    fprintf(stderr, "latency: expecting a number\n");
    exit(1);
  }

  x = atoi(s);
  ret = x;

  for(i = 0; s[i]; i++){
    if(isdigit(s[i])){
      /* ok */
    } else if(s[i] == 'k' || s[i] == 'K'){
      ret = x * (long long) 1024;
    } else if(s[i] == 'm' || s[i] == 'M'){
      ret = x * (long long) 1024 * (long long) 1024;
    } else {
      fprintf(stderr, "latency: could not parse number '%s'\n", s);
      exit(1);
    }
  }

  x = ret;
  if(ret != x){
    fprintf(stderr, "latency: argument %s was too large\n", s);
    exit(1);
  }

  return(x);
}
